
//		PopInfo.cpp


#define UNICODE							1
#define _UNICODE						1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <AclAPI.h>
#include <Shlwapi.h>

#include "PlugIn.h"
#include "TakeCmd.h"

#include "resource.h"


#define VER_MAJOR						1
#define VER_MINOR						4
#define VER_BUILD						0
#define VER_PATCH						4

#define PLUGIN_NAME						L"PopInfo"

#ifdef _WIN64
#define PLATFORM						L"x64"
#else
#define PLATFORM						L"x86"
#endif

#define MAX_BUFFER_LEN					4096
#define MAX_ARG_LEN						8192
#define MAX_FILENAME_LEN				32768
#define SHORT_BUF_LEN					1024

#define ERRHANDLE						GetStdHandle( STD_ERROR_HANDLE )

#define SAFE_PERCENT					0xFF05


const wchar_t PluginRegKey[]			= L"Software\\JPPlugins\\PopInfo";
const wchar_t HotkeyRegValue[]			= L"Hotkey";
const wchar_t SmartTimesRegValue[]		= L"SmartTimes";
const wchar_t SubdirCountsRegValue[]	= L"SubdirCounts";

const wchar_t OnOff[2][4]				= { L"OFF", L"ON" };
const wchar_t TimeOpts[3][4]			= { L"OFF", L"ON", L"UTC" };


int TriggerKey							= 0x11;
int LocalTimes							= 1;
BOOL SubdirCounts						= TRUE;


const int MaxArgs						= 128;
const int MaxArgLen						= 2048;
wchar_t arg[MaxArgs][MaxArgLen + 1];
int arglen[MaxArgs];


HINSTANCE MyDLLHandle					= NULL;


struct _PopInfoData {
	LPTSTR Title;
	LPTSTR FullName;
	LPTSTR ShortName;
	LPTSTR AltName;
	LPTSTR Description;
	LPTSTR Size;
	LPTSTR Attributes;
	LPTSTR CreatedStamp;
	LPTSTR ModifiedStamp;
	LPTSTR AccessedStamp;
	LPTSTR Owner;
	LPTSTR AssocStr;
	LPTSTR OpensWith;
	int AssocType;
	DWORD AttribsNumeric;
	unsigned int Version[4];
};




wchar_t PopInfoHelpText[] = L"Display or change options for the PopInfo plugin. \n\
\n\
POPINFO [/S /R /L /T:x /C:x] [hotkey] \r\n\n\
\t/S \tsave options to the registry as new default  \n\
\t/R \trestore saved options from the registry\n\
\t/L \tlist all supported hotkeys\n\
\t/T \tsmart timestamps:     0: off, 1: on, 2: UTC\n\
\t/C \tsubdirectory counts:  0: off, 1: on\n\
\n\
Examples of hotkeys are Control-Q, Alt-I, F2, or Shift-F10.  \n\
Control-Q is the default.  If no hotkey is specified, the\n\
current setting will be shown.\n";



#include "HotKeys.cpp"
#include "ForceFGWin.cpp"




BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:

			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



int UnsupportedVersion( void )
{
	wchar_t TempBuffer[SHORT_BUF_LEN]	= L"%_4ver";
	wchar_t *sep1						= NULL;
	long int ShellMajor = -1, ShellMinor = -1, ShellBuild = -1;

	ExpandVariables( TempBuffer, 1 );
	ShellMajor = wcstol( TempBuffer, &sep1, 10 );

	if ( *sep1 ) {
		sep1++;
		if ( *sep1 )
			ShellMinor = wcstol( sep1, NULL, 10 );
	}

	wcscpy_s( TempBuffer, SHORT_BUF_LEN, L"%_build" );
	ExpandVariables( TempBuffer, 1 );
	ShellBuild = wcstol( TempBuffer, NULL, 10 );

	if ( ( ShellMajor < 7 ) || ( ShellMajor > 99 ) || ( ShellMinor < 0 ) || ( ShellMinor > 99 ) || ( ShellBuild < 0 ) || ( ShellBuild > 9999 ) )
		return -1;			//  version number fails sanity checks

	if ( ShellMajor < 11 )
		return 1;
	else if ( ShellMajor > 11 )
		return 0;

	if ( ShellMinor > 0 )
		return 0;

	if ( ShellBuild < 48 )
		return 1;

	return 0;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll         = (LPTSTR)L"PopInfo";
	piInfo.pszAuthor      = (LPTSTR)L"Charles Dye";
	piInfo.pszEmail       = (LPTSTR)L"cdye@unm.edu";
	piInfo.pszWWW         = (LPTSTR)L"http://charlesdye.net/plugins/popinfo.html";
	piInfo.pszDescription = (LPTSTR)L"PopInfo";
	piInfo.pszFunctions   = (LPTSTR)L"*key,popinfo";
	piInfo.nMajor         = VER_MAJOR;
	piInfo.nMinor         = VER_MINOR;
	piInfo.nBuild         = VER_BUILD;

	return &piInfo;
}



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_BUFFER_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables(shortbuf, 1);
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s(shortbuf, MAX_BUFFER_LEN, L"%_batchname [%_batchline]  ");
	ExpandVariables(shortbuf, 1);
	Qprintf( ERRHANDLE, shortbuf);

	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int SaveOptions( BOOL ComplainOnError )
{
	int rv = 0;
	HKEY KeyHandle = NULL;

	if ( RegCreateKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, NULL, 0, KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( RegSetValueEx( KeyHandle, HotkeyRegValue, 0, REG_DWORD, (LPBYTE) &TriggerKey, 4 ) == ERROR_SUCCESS ) {
			RegSetValueEx( KeyHandle, SmartTimesRegValue, 0, REG_DWORD, (LPBYTE) &LocalTimes, 4 );
			RegSetValueEx( KeyHandle, SubdirCountsRegValue, 0, REG_DWORD, (LPBYTE) &SubdirCounts, 4 );
		}
		else {
			if ( ComplainOnError ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error writing to registry\r\n" );
			}
			rv = -1;
		}

		RegCloseKey( KeyHandle );
	}
	else {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening registry key\r\n" );
		}
		rv = -1;
	}

	return rv;
}



int RestoreOptions( BOOL ComplainOnError )
{
	int rv = 0;
	HKEY KeyHandle = NULL;
	DWORD keyval = 0xffff0000, smartval = 0xffff0000, countval = 0xffff0000;
	DWORD buflen = 0;
	

	if ( RegOpenKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		buflen = 4;
		if ( RegQueryValueEx( KeyHandle, HotkeyRegValue, NULL, NULL, (LPBYTE) &keyval, &buflen ) == ERROR_SUCCESS ) {
			buflen = 4;
			RegQueryValueEx( KeyHandle, SmartTimesRegValue, NULL, NULL, (LPBYTE) &smartval, &buflen );
			buflen = 4;
			RegQueryValueEx( KeyHandle, SubdirCountsRegValue, NULL, NULL, (LPBYTE) &countval, &buflen );
		}
		else
			rv = -1;

		RegCloseKey( KeyHandle );
	}
	else
		rv = -1;

	if ( rv ) {
		if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, PluginRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

			buflen = 4;
			if ( RegQueryValueEx( KeyHandle, HotkeyRegValue, NULL, NULL, (LPBYTE) &keyval, &buflen ) == ERROR_SUCCESS ) {
				buflen = 4;
				RegQueryValueEx( KeyHandle, SmartTimesRegValue, NULL, NULL, (LPBYTE) &smartval, &buflen );
				buflen = 4;
				RegQueryValueEx( KeyHandle, SubdirCountsRegValue, NULL, NULL, (LPBYTE) &countval, &buflen );
				rv = 0;
			}

			RegCloseKey( KeyHandle );
		}
	}

	if ( rv ) {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error reading from registry\r\n" );
		}

		return rv;
	}

	if ( !IsKnownKeyValue( keyval ) ) {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid hotkey value in registry\r\n" );
		}
		return -2;
	}

	TriggerKey   = keyval;
	SubdirCounts = ( countval != 0 );

	if ( ( smartval >= 0 ) && ( smartval <= 2 ) )
		LocalTimes = smartval;
	else
		LocalTimes = 1;

	return 0;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	InitHotkeys();

	if ( UnsupportedVersion() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"TCC v11.00.48 or higher required!\r\n" );
		return -1;
	}

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";

	if ( RestoreOptions( FALSE ) ) {
		TriggerKey   = 0x11;
		LocalTimes   = 1;
		SubdirCounts = TRUE;
	}

	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) ) {
		GetKeyName( TriggerKey, shortbuf );
		Printf( L"PopInfo v%d.%d.%d loaded; hotkey is %s.\n", VER_MAJOR, VER_MINOR, VER_BUILD, shortbuf );
	}

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



int DivideCommandArgs( LPTSTR lpszString )
{
	int argnum = 0;
	int i = 0;
	int a = 0;
	int n = 0;
	wchar_t ch;
	BOOL quote = FALSE;
	BOOL addme = FALSE;
	BOOL done = FALSE;

	for ( i = 0; i < MaxArgs; i++ ) {
		arg[i][0] = '\0';
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = lpszString[i];
		addme = TRUE;

		if ( ch == '"' ) {
			quote = !quote;
			addme = FALSE;
		}

		if ( iswspace(ch) && !quote ) {
			addme = FALSE;
			if ( a > 0 ) {
				argnum++;
				a = 0;
			}
		}

		if ( (ch == '/') && !quote && (a > 0) && (arg[argnum][0] == '/') ) {
			argnum++;
			a = 0;
		}

		if ( ch == '\0' ) {
			argnum++;
			a = 0;
			addme = FALSE;
			done = TRUE;
		}

		if ( addme && (argnum < MaxArgs) && (a < MaxArgLen) ) {
			arg[argnum][a] = towupper(ch);
			a++;
			arg[argnum][a] = '\0';
			arglen[argnum] = a;
		}
	}

	if ( argnum > MaxArgs )
		argnum = MaxArgs;

	return argnum;
}



int ContainsWildcards( LPTSTR inString )
{
	int i = 0;
	wchar_t ch = '\0';

	while ( inString[i] != '\0' ) {
		ch = inString[i];

		if ( ch == '*' )
			return 1;

		if ( ch == '?' ) {
			if ( i != 2 )
				return 1;

			if ( ( inString[0] != '\\' ) || ( inString[1] != '\\' ) || ( inString[3] != '\\' ) )
				return 1;
		}

		i++;
	}

	return 0;
}



int GetBaseName( LPTSTR inString, LPTSTR outString )
{
	int i = 0, j = 0;
	outString[0] = '\0';
	wchar_t ch = '\0';

	do {
		ch = inString[i++];

		if ( ch == '\\' ) {
			j = 0;
			outString[j] = '\0';
		}
		else
			outString[j++] = towupper(ch);

	} while ( ch != 0 );

	return 0;
}



LPTSTR _ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



int FixDoubledPercents( LPTSTR inString )
{
	int rv = 0;

	int i = 0, j = 0;
	wchar_t ch = '\0';

	do {
		ch = inString[i];

		if ( ch == '%' )
			if ( inString[i + 1] == '%' ) {

				j = i;
				do {
					inString[j] = inString[j + 1];
					j++;
				} while ( inString[j] != '\0' );

				rv++;
			}

		i++;
	} while ( ch != '\0' );

	return rv;
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



int ReplaceChar( LPTSTR String, wchar_t From, wchar_t To )
{
	if ( !String )
		return -1;


	size_t i		= 0;
	int Count		= 0;


	while ( String[i] ) {
		if ( String[i] == From ) {
			String[i] = To;
			Count++;
		}

		i++;
	}
	
	return Count;
}



int UndoubleChar( LPTSTR String, wchar_t Char )
{
	if ( !String )
		return -1;

	if ( ( String[0] == '\0' ) || ( Char == '\0' ) )
		return 0;


	size_t i = 0, j = 0;
	int Count		= 0;
	wchar_t ch		= '\0';


	do {
		ch = String[i++];
		String[j++] = ch;
		if ( ( ch == Char ) && ( String[i] == Char ) )
			i++, Count++;
	}
	while ( ch );

	return Count;
}



int DecodeFileURI( LPTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t temp[MAX_FILENAME_LEN]			= L"";
	wchar_t LocalName[MAX_FILENAME_LEN]		= L"";
	DWORD LocalNameLen						= MAX_ARG_LEN;
	int rv									= 0;


	wcscpy_s( temp, MAX_FILENAME_LEN, Filename );
	ReplaceChar( temp, SAFE_PERCENT, '%' );
	LocalNameLen = MAX_FILENAME_LEN;
	LocalName[0] = '\0';
	rv = PathCreateFromUrl( temp, LocalName, &LocalNameLen, NULL );

	if ( ( rv == S_OK ) && LocalName[0] ) {
		if ( QueryIsFileOrDirectory( LocalName ) ) {
			wcscpy_s( Filename, MAX_FILENAME_LEN, LocalName );
			return 1;
		}
	}

	wcscpy_s( temp, MAX_FILENAME_LEN, Filename );
	ReplaceChar( temp, SAFE_PERCENT, '%' );
	LocalNameLen = MAX_FILENAME_LEN;
	LocalName[0] = '\0';

	if ( UndoubleChar( temp, '%' ) > 0 ) {
		rv = PathCreateFromUrl( temp, LocalName, &LocalNameLen, NULL );

		if ( ( rv == S_OK ) && LocalName[0] ) {
			if ( QueryIsFileOrDirectory( LocalName ) ) {
				wcscpy_s( Filename, MAX_FILENAME_LEN, LocalName );
				return 1;
			}
		}
	}

	return 0;
}



int GetFileDescription( LPTSTR Filename, LPTSTR Description, unsigned int BufferLen )
{
	if ( !Filename || !Description )
		return -1;

	if ( Filename[0] == '\0' )
		return ERROR_FILE_NOT_FOUND;

	if ( !QueryIsFileOrDirectory( Filename ) )
		return ERROR_FILE_NOT_FOUND;


	wchar_t temp[MAX_ARG_LEN]		= L"%@descript[";


	wcscat_s( temp, MAX_ARG_LEN, Filename );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	ExpandVariables( temp, 1 );

	if ( (unsigned int) wcslen( temp ) >= BufferLen )
		return ERROR_BUFFER_OVERFLOW;

	if ( temp[0] )
		wcscpy_s( Description, BufferLen, temp );
	else
		wcscpy_s( Description, BufferLen, L"\u2014" );

	return 0;
}



int MakePrettyFileSize( LPTSTR outString, __int64 FileSize )
{
	wchar_t shortbuf[SHORT_BUF_LEN];
	double bigunits;
	wchar_t unitname[16]		= L"" ;


	wsprintf( shortbuf, L"%I64u", FileSize );
	InsertCommas( shortbuf, SHORT_BUF_LEN );
	wcscpy_s( outString, SHORT_BUF_LEN, shortbuf );
	wcscat_s( outString, SHORT_BUF_LEN, L" bytes" );

	if ( FileSize >= 1073741824 ) {
		bigunits = FileSize / 1073741824.0;
		wcscpy_s( unitname, 16, L"GB" );
	}
	else if ( FileSize >= 1048576 ) {
		bigunits = FileSize / 1048576.0;
		wcscpy_s( unitname, 16, L"MB" );
	}
	else if ( FileSize >= 1024 ) {
		bigunits = FileSize / 1024.0;
		wcscpy_s( unitname, 16, L"KB" );
	}

	if ( unitname[0] ) {
		swprintf( shortbuf, SHORT_BUF_LEN, L"%.2f", bigunits );
		InsertCommas( shortbuf, SHORT_BUF_LEN );

		wcscat_s( outString, SHORT_BUF_LEN, L"  (" );
		wcscat_s( outString, SHORT_BUF_LEN, shortbuf );
		wcscat_s( outString, SHORT_BUF_LEN, L" " );
		wcscat_s( outString, SHORT_BUF_LEN, unitname );
		wcscat_s( outString, SHORT_BUF_LEN, L")" );
	}

	return 0;
}



int MakeSubdirCounts( LPTSTR outString, LPTSTR DirName )
{

	long int nFiles = 0, nSubdirs = 0;
	int rv = 0;
	wchar_t FileSpec[MAX_FILENAME_LEN]	= L"";
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";

	WIN32_FIND_DATA FileInfo;
	HANDLE FindHandle = NULL;

	wcscpy_s( FileSpec, MAX_FILENAME_LEN, DirName );
	wcscat_s( FileSpec, MAX_FILENAME_LEN, L"\\*" );

	FindHandle = FindFirstFile( FileSpec, &FileInfo );

	if ( FindHandle != INVALID_HANDLE_VALUE ) {

		do {

			if ( FileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
				if ( wcscmp( FileInfo.cFileName, L"." ) && wcscmp( FileInfo.cFileName, L".." ) )
					nSubdirs++;
			}
			else
				nFiles++;
			
		} while ( FindNextFile( FindHandle, &FileInfo ) != 0 );

		FindClose( FindHandle );
	}

	wsprintf( shortbuf, L"%d", nFiles );
	InsertCommas( shortbuf, SHORT_BUF_LEN );
	wcscpy_s( outString, SHORT_BUF_LEN, shortbuf );

	if ( nFiles == 1 )
		wcscat_s( outString, SHORT_BUF_LEN, L" file, " );
	else
		wcscat_s( outString, SHORT_BUF_LEN, L" files, " );

	wsprintf( shortbuf, L"%d", nSubdirs );
	InsertCommas( shortbuf, SHORT_BUF_LEN );
	wcscat_s( outString, SHORT_BUF_LEN, shortbuf );

	if ( nSubdirs == 1 )
		wcscat_s( outString, SHORT_BUF_LEN, L" subdir" );
	else
		wcscat_s( outString, SHORT_BUF_LEN, L" subdirs" );

	return 0;
}



int MakeFileAttributes( LPTSTR outString, DWORD Attribs )
{
	if ( Attribs & FILE_ATTRIBUTE_READONLY )
		wcscat_s( outString, SHORT_BUF_LEN, L"ReadOnly " );

	if ( Attribs & FILE_ATTRIBUTE_HIDDEN )
		wcscat_s( outString, SHORT_BUF_LEN, L"Hidden " );

	if ( Attribs & FILE_ATTRIBUTE_SYSTEM )
		wcscat_s( outString, SHORT_BUF_LEN, L"System " );

	if ( Attribs & FILE_ATTRIBUTE_ARCHIVE )
		wcscat_s( outString, SHORT_BUF_LEN, L"Archive " );

	if ( Attribs & FILE_ATTRIBUTE_DIRECTORY )
		wcscat_s( outString, SHORT_BUF_LEN, L"Dir " );

	if ( Attribs & FILE_ATTRIBUTE_ENCRYPTED )
		wcscat_s( outString, SHORT_BUF_LEN, L"Encrypt " );

	if ( Attribs & FILE_ATTRIBUTE_COMPRESSED )
		wcscat_s( outString, SHORT_BUF_LEN, L"Compress " );

	if ( Attribs & FILE_ATTRIBUTE_NOT_CONTENT_INDEXED )
		wcscat_s( outString, SHORT_BUF_LEN, L"noIndex " );

	if ( Attribs & FILE_ATTRIBUTE_REPARSE_POINT )
		wcscat_s( outString, SHORT_BUF_LEN, L"Junction " );

	if ( Attribs & FILE_ATTRIBUTE_OFFLINE )
		wcscat_s( outString, SHORT_BUF_LEN, L"Offline " );

	if ( Attribs & FILE_ATTRIBUTE_SPARSE_FILE )
		wcscat_s( outString, SHORT_BUF_LEN, L"Sparse " );

	if ( Attribs & FILE_ATTRIBUTE_TEMPORARY )
		wcscat_s( outString, SHORT_BUF_LEN, L"Temp " );

	if ( ( Attribs & ~FILE_ATTRIBUTE_NORMAL ) == 0 )
		wcscat_s( outString, SHORT_BUF_LEN, L"- none -" );

	return 0;
}



int MakeStamp( LPTSTR outString, FILETIME inTime, BOOL SuppressTime, BOOL BeSmart )
{
	SYSTEMTIME FileStamp, FileStampLocal;
	FILETIME FileTimeLocal;
	wchar_t shortbuf[SHORT_BUF_LEN];


	if ( LocalTimes == 2 ) {
		FileTimeToSystemTime( &inTime, &FileStampLocal );

		if ( SuppressTime )
			wsprintf( outString, L"%04d-%02d-%02d UTC", FileStampLocal.wYear, FileStampLocal.wMonth, FileStampLocal.wDay );
		else
			wsprintf( outString, L"%04d-%02d-%02d  %02d:%02d:%02d UTC", FileStampLocal.wYear, FileStampLocal.wMonth, FileStampLocal.wDay, FileStampLocal.wHour, FileStampLocal.wMinute, FileStampLocal.wSecond );

		return 0;
	}

	if ( BeSmart ) {
		FileTimeToSystemTime( &inTime, &FileStamp );
		SystemTimeToTzSpecificLocalTime( NULL, &FileStamp, &FileStampLocal );
	}
	else {
		FileTimeToLocalFileTime( &inTime, &FileTimeLocal );
		FileTimeToSystemTime( &FileTimeLocal, &FileStampLocal );
	}


	GetDateFormat( LOCALE_USER_DEFAULT, DATE_LONGDATE, &FileStampLocal, NULL, outString, SHORT_BUF_LEN );

	if ( !SuppressTime ) {
		wcscat_s( outString, SHORT_BUF_LEN, L"  " );
		GetTimeFormat( LOCALE_USER_DEFAULT, 0, &FileStampLocal, NULL, shortbuf, SHORT_BUF_LEN );
		wcscat_s( outString, SHORT_BUF_LEN, shortbuf );
	}

	return 0;
}



int MakeOwner( LPTSTR outString, LPTSTR Filename, size_t buflen )
{
	if ( !outString || !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	HANDLE FileHandle							= NULL;
	PSECURITY_DESCRIPTOR SecurityDescriptor		= NULL;
	PSID OwnerSid								= NULL;
	SID_NAME_USE SidType;
	wchar_t OwnerName[SHORT_BUF_LEN]			= L"";
	wchar_t DomainName[SHORT_BUF_LEN]			= L"";
	DWORD OwnerBufLen							= SHORT_BUF_LEN;
	DWORD DomainBufLen							= SHORT_BUF_LEN;


	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL );
	if ( !FileHandle || ( FileHandle == INVALID_HANDLE_VALUE ) )
		return 1;

	if ( GetSecurityInfo( FileHandle, SE_FILE_OBJECT, OWNER_SECURITY_INFORMATION, &OwnerSid, NULL, NULL, NULL, &SecurityDescriptor ) == ERROR_SUCCESS ) {
		if ( SecurityDescriptor ) {
			if ( OwnerSid )
				LookupAccountSid( NULL, OwnerSid, OwnerName, &OwnerBufLen, DomainName, &DomainBufLen, &SidType );

			LocalFree( SecurityDescriptor );
		}
	}

	CloseHandle( FileHandle );

	if ( OwnerName[0] ) {

		if ( DomainName[0] ) {
			wcscpy_s( outString, buflen, DomainName );
			wcscat_s( outString, buflen, L"\\" );
		}
		else
			outString[0] = '\0';

		wcscat_s( outString, buflen, OwnerName );
	}
	
	return 0;
}



int GetAssocInfo( LPTSTR inFilename, LPTSTR outFileType, LPTSTR outOpensWith, int *outKey )
{
	if ( !inFilename || !outFileType || !outOpensWith || !outKey )
		return -1;


	outFileType[0]	= '\0';
	outOpensWith[0]	= '\0';
	*outKey			= 0;


	if ( QueryIsDirectory( inFilename ) ) {
		*outKey = 10;
		return 0;
	}

	wchar_t FileExt[MAX_BUFFER_LEN] = L"";
	int i = 0, j = 0;
	wchar_t ch = '\0';

	do {
		ch = inFilename[i++];

		if ( ( ch == '\\' ) || ( ch == ':' ) ) {
			j = 0;
			FileExt[0] = '\0';
		}
		else if ( ch == '.' ) {
			FileExt[0] = ch;
			FileExt[1] = '\0';
			j = 1;
		}
		else
			FileExt[j++] = towupper(ch);

	} while ( ch != '\0' );


	if ( FileExt[0] != '.' )
		return -1;

	if ( FileExt[1] == '\0' )
		return -1;


	wchar_t KeyName[MAX_BUFFER_LEN]			= L"";
	wchar_t FileType[MAX_BUFFER_LEN]		= L"";
	wchar_t OpenString[MAX_BUFFER_LEN]		= L"";
	wchar_t PrettyAssoc[MAX_BUFFER_LEN]		= L"";
	DWORD BufLen = 0;
	int KeyFound = 0;
	HKEY KeyHandle = NULL;


	wcscpy_s( KeyName, MAX_BUFFER_LEN, L"Software\\Classes\\" );
	wcscat_s( KeyName, MAX_BUFFER_LEN, FileExt );


	if ( RegOpenKeyEx( HKEY_CURRENT_USER, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		FileType[0] = '\0';

		BufLen = sizeof( FileType );
		if ( RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) FileType, &BufLen ) == ERROR_SUCCESS )
			if ( FileType[0] )
				KeyFound = 1;

		RegCloseKey( KeyHandle );
	}


	if ( !KeyFound ) {
		
		if ( RegOpenKeyEx( HKEY_CLASSES_ROOT, FileExt, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
			FileType[0] = '\0';

			BufLen = sizeof( FileType );
			if ( RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) FileType, &BufLen ) == ERROR_SUCCESS )
				if ( FileType[0] )
					KeyFound = 2;

			RegCloseKey( KeyHandle );
		}
	}


	if ( KeyFound ) {
		wcscpy_s( KeyName, MAX_BUFFER_LEN, L"Software\\Classes\\" );
		wcscat_s( KeyName, MAX_BUFFER_LEN, FileType );

		//  try to read pretty filetype from user hive:

		if ( RegOpenKeyEx( HKEY_CURRENT_USER, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
			BufLen = sizeof(PrettyAssoc);
			RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) PrettyAssoc, &BufLen );
			RegCloseKey( KeyHandle );
		}

		//  if that didn't work, try to read pretty filetype from system hive:

		if ( PrettyAssoc[0] == '\0' )
			if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
				BufLen = sizeof(PrettyAssoc);
				RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) PrettyAssoc, &BufLen );
				RegCloseKey( KeyHandle );
			}

		//  try to read open string from user hive:

		wcscat_s( KeyName, MAX_BUFFER_LEN, L"\\shell\\open\\command" );

		if ( RegOpenKeyEx( HKEY_CURRENT_USER, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
			BufLen = sizeof(OpenString);
			RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) OpenString, &BufLen );
			RegCloseKey( KeyHandle );
		}

		//  if that didn't work, try to read open string from system hive:

		if ( OpenString[0] == '\0' )
			if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, KeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
				BufLen = sizeof(OpenString);
				RegQueryValueEx( KeyHandle, NULL, NULL, NULL, (LPBYTE) OpenString, &BufLen );
				RegCloseKey( KeyHandle );
			}
	}

	if ( OpenString[0] == '\0' )
		wcscpy_s( OpenString, MAX_BUFFER_LEN, L"\u2014" );
	else {
		int i = 0, j = 0;
		wchar_t ch = '\0';
		BOOL quote = FALSE;

		do {
			ch = OpenString[i++];

			if ( ch == '"' )
				quote = !quote;
			else {
				if ( iswspace(ch) && ( j > 0 ) && !quote )
					ch = '\0';

				if ( iswspace(ch) && ( j == 0 ) && !quote )
					continue;

				KeyName[j++] = ch;
			}
		} while ( ch != '\0' );

		if ( KeyName[0] == '%' )
			if ( ( KeyName[1] == '1' ) || ( towupper(KeyName[1]) == 'L' ) )
				wcscpy_s( OpenString, MAX_BUFFER_LEN, L"Executable file" );
	}


	if ( KeyFound ) {
		wcscpy_s( outFileType, MAX_FILENAME_LEN, FileType );

		if ( PrettyAssoc[0] ) {
			wcscat_s( outFileType, MAX_FILENAME_LEN, L"  (" );
			wcscat_s( outFileType, MAX_FILENAME_LEN, PrettyAssoc );
			wcscat_s( outFileType, MAX_FILENAME_LEN, L")" );
		}

		wcscpy_s( outOpensWith, MAX_FILENAME_LEN, OpenString ); 
	}

	*outKey = KeyFound;
	return 0;
}



int GetFileVersion( LPTSTR Filename, unsigned int *Version )
{
	DWORD VerInfoSize				= 0;
	void *Buffer					= NULL;
	VS_FIXEDFILEINFO *FileInfo		= NULL;
	UINT FileInfoSize				= 0;

	int rv							= -1;


	if ( !QueryIsFile( Filename ) )
		return ERROR_FILE_NOT_FOUND;


	VerInfoSize = GetFileVersionInfoSize( Filename, NULL );

	if ( VerInfoSize ) {
		Buffer = malloc( VerInfoSize );

		if ( Buffer ) {
			if ( GetFileVersionInfo( Filename, 0, VerInfoSize, Buffer ) ) {
				VerQueryValue( Buffer, L"\\", (void**) &FileInfo, &FileInfoSize );

				if ( FileInfo ) {
					Version[0]		= FileInfo->dwFileVersionMS >> 16;
					Version[1]		= FileInfo->dwFileVersionMS & 0xffff;
					Version[2]		= FileInfo->dwFileVersionLS >> 16;
					Version[3]		= FileInfo->dwFileVersionLS & 0xffff;
					rv = 0;
				}
			}

			free( Buffer );
		}
		else
			rv = ERROR_NOT_ENOUGH_MEMORY;
	}

	return rv;
}



/*
int MakeWordCase( LPTSTR inString, BOOL SmallExtension )
{
	BOOL WordStart = TRUE;
	int i = 0;
	wchar_t ch = '\0', ch2 = '\0';
	int lastdot = -1;


	if ( SmallExtension )
		for ( i = 0; inString[i] != '\0'; i++ )
			if ( inString[i] == '.' )
				lastdot = i;

	i = 0;
	do {
		ch = inString[i];

		if ( WordStart )
			ch = towupper( ch );
		else
			ch = towlower( ch );

		if ( ch == '\'' ) {
			if ( i == 0 )
				WordStart = TRUE;
			else {
				WordStart = FALSE;
				ch2 = towupper( inString[i - 1] );
				if ( ( ch2 == 'O' ) || (  ch2 == 'M' ) ) {
					WordStart = TRUE;

					if ( i > 1 )
						if ( iswalpha( inString[i - 2] ) )
							WordStart = FALSE;
				}
			}
		}
		else
			WordStart = !iswalpha( ch );

		if ( SmallExtension )
			if ( i >= lastdot )
				WordStart = FALSE;

		inString[i] = ch;
		i++;
	} while ( ch != '\0' );
	
	return 0;
}
*/



int GetVolInfo( LPTSTR inString, LPTSTR outFSType )
{

	if ( !inString || !outFSType )
		return -1;

	wchar_t tempbuf[MAX_BUFFER_LEN]		= L"";
	wchar_t volname[SHORT_BUF_LEN]		= L"";
	int numBackslashes = 4;
	int i = 0, j = 0, k = 0;
	wchar_t ch = '\0';

	if ( iswalpha(inString[0]) && ( inString[1] == ':' ) && ( inString[2] == '\\' ) )
		numBackslashes = 1;
	else if ( ( inString[0] != '\\' ) || ( inString[1] != '\\' ) )
		return -1;

	do {
		ch = inString[i++];

		if ( k == numBackslashes )
			ch = '\0';

		if ( ch == '\\' )
			k++;

		tempbuf[j++] = ch;
	} while ( ch != '\0' );

	outFSType[0] = '\0';

	if ( !GetVolumeInformation( tempbuf, volname, SHORT_BUF_LEN, NULL, NULL, NULL, outFSType, MAX_BUFFER_LEN ) )
		return -1;

	if ( outFSType[0] == '\0' )
		return -1;

	return 0;
}



INT_PTR CALLBACK DialogProc$( HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam )
{
	switch ( Message ) {

			case WM_INITDIALOG :

				{
					wchar_t temp[SHORT_BUF_LEN]		= L"";

					_PopInfoData *Data = (_PopInfoData*) lParam;


					if ( Data->Title )
						SetWindowText( hWnd, Data->Title );

					if ( Data->FullName )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT1 ), Data->FullName );

					if ( Data->AltName )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT2 ), Data->AltName );

					if ( Data->Description )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT3 ), Data->Description );

					if ( Data->Version[0] || Data->Version[1] || Data->Version[2] || Data->Version[3] ) {
						wsprintf( temp, L"%u.%u.%u.%u", Data->Version[0], Data->Version[1], Data->Version[2], Data->Version[3] );
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT4 ), temp );
					}

					if ( ( Data->AttribsNumeric ) & FILE_ATTRIBUTE_DIRECTORY )
						SetWindowText( GetDlgItem( hWnd, IDC_STATIC5 ), L"Contains:" );

					if ( Data->Size )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT5 ), Data->Size );

					if ( Data->Attributes )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT6 ), Data->Attributes );

					if ( Data->CreatedStamp )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT7 ), Data->CreatedStamp );

					if ( Data->ModifiedStamp )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT8 ), Data->ModifiedStamp );

					if ( Data->AccessedStamp )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT9 ), Data->AccessedStamp );

					if ( Data->Owner )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT10 ), Data->Owner );

					if ( Data->AssocType == 1 )
						SetWindowText( GetDlgItem( hWnd, IDC_STATIC11 ), L"User file type:" );

					if ( Data->AssocStr )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT11 ), Data->AssocStr );

					if ( Data->OpensWith )
						SetWindowText( GetDlgItem( hWnd, IDC_EDIT12 ), Data->OpensWith );


					ForceForegroundWindow( hWnd );
					return FALSE;
				}

			case WM_COMMAND :

				switch ( LOWORD( wParam ) ) {

				case IDOK :
				case IDCANCEL :

					EndDialog( hWnd, LOWORD( wParam ) );
					break;

				default :
					//	wprintf( L"   * WM_COMMAND:  wParam = 0x%08X\n", wParam );
					break;
				}

				return FALSE;

			case WM_DESTROY :

				return FALSE;

			default :

				return FALSE;
	}
}



int ShowFileInfo( LPTSTR Filename )
{

	if ( !Filename ) {
		honk();
		return -1;
	}
	else if ( Filename[0] == '\0' ) {
		honk();
		return -2;
	}


	wchar_t TempFilename[MAX_FILENAME_LEN]	= L"";
	wchar_t Truename[MAX_FILENAME_LEN]		= L"";
	wchar_t ShortName[MAX_FILENAME_LEN]		= L"";
	wchar_t Ext[SHORT_BUF_LEN]				= L"";
	wchar_t fstype[SHORT_BUF_LEN]			= L"";
	wchar_t Title[SHORT_BUF_LEN]			= L"";
	wchar_t FileDescription[MAX_ARG_LEN]	= L"";
	wchar_t FileSizeStr[SHORT_BUF_LEN]		= L"";
	wchar_t Attributes[SHORT_BUF_LEN]		= L"";
	wchar_t CreatedStamp[SHORT_BUF_LEN]		= L"";
	wchar_t ModifiedStamp[SHORT_BUF_LEN]	= L"";
	wchar_t AccessedStamp[SHORT_BUF_LEN]	= L"";
	wchar_t Owner[SHORT_BUF_LEN]			= L"";
	wchar_t AssocStr[MAX_FILENAME_LEN]		= L"";
	wchar_t OpensWith[MAX_FILENAME_LEN]		= L"";

	size_t i				= 0;
	BOOL VolumeIsFAT		= FALSE;
	BOOL VolumeIsNTFS		= FALSE;
	int AssocType			= 0;

	_PopInfoData Data;


	wcscpy_s( TempFilename, MAX_FILENAME_LEN, Filename );


	//		Check for and decode file:/// URIs :

	if ( !_wcsnicmp( TempFilename, L"file://", 7 ) )
		DecodeFileURI( TempFilename );


	//		Canonicalize filename :

	Truename[0] = '\0';
	QueryTrueName( TempFilename, Truename );
	if ( Truename[0] == '\0' ) {
		honk();
		return -3;
	}

	i = wcslen( Truename );
	if ( i )
		if ( Truename[i - 1] == '\\' )
			Truename[i - 1] = '\0';


	//		Extract the filename from the canonicalized name :

	GetBaseName( Truename, ShortName );
	_ExtensionPart( Truename, Ext );


	//		Check for wildcards :

	if ( ContainsWildcards( Truename ) ) {
		honk();
		Sleep( 120 );
		honk();
		return -4;
	}


	//		Get volume filesystem type :

	if ( !GetVolInfo( Truename, fstype ) ) {
		VolumeIsFAT  = ( !_wcsnicmp( fstype, L"FAT", 3 ) );
		VolumeIsNTFS = ( !_wcsnicmp( fstype, L"NTFS", 4 ) );
	}


	//		Get file info :

	WIN32_FIND_DATA FileInfo;
	
	HANDLE FindHandle = FindFirstFile( Truename, &FileInfo );
	
	if ( FindHandle == INVALID_HANDLE_VALUE ) {

		if ( !FixDoubledPercents( TempFilename ) ) {
			honk();
			return -16;
		}
		else {
			Truename[0] = '\0';
			QueryTrueName( TempFilename, Truename );
			if ( Truename[0] == '\0' ) {
				honk();
				return -3;
			}

			i = wcslen( Truename );
			if ( i > 0 )
				if ( Truename[i - 1] == '\\' )
					Truename[i - 1] = '\0';

			GetBaseName( Truename, ShortName );

			FindHandle = FindFirstFile( Truename, &FileInfo );

			if ( FindHandle == INVALID_HANDLE_VALUE ) {
				honk();
				return -16;
			}
		}
	}

	FindClose( FindHandle );


	//		Create dialog title :

	wsprintf( Title, L"%s \u2014 %s %u.%u.%u", ShortName, PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );


	//		File description :

	GetFileDescription( Truename, FileDescription, MAX_ARG_LEN );


	//		File size :

	if ( FileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
		if ( SubdirCounts )
			MakeSubdirCounts( FileSizeStr, Truename );
	}
	else
		MakePrettyFileSize( FileSizeStr, FileInfo.nFileSizeLow | ( (unsigned __int64) FileInfo.nFileSizeHigh << 32 ) );


	//		Attributes :

	MakeFileAttributes( Attributes, FileInfo.dwFileAttributes );


	//		Time stamps :

	MakeStamp( CreatedStamp, FileInfo.ftCreationTime, FALSE, VolumeIsNTFS );
	MakeStamp( ModifiedStamp, FileInfo.ftLastWriteTime, FALSE, VolumeIsNTFS );
	MakeStamp( AccessedStamp, FileInfo.ftLastAccessTime, VolumeIsFAT, VolumeIsNTFS );


	//		Owner :

	MakeOwner( Owner, Truename, SHORT_BUF_LEN );


	//		File association :

	GetAssocInfo( Truename, AssocStr, OpensWith, &AssocType );


	//		Fill in the Data structure :

	memset( &Data, 0x00, sizeof( _PopInfoData ) );
	Data.Title			= Title;
	Data.FullName		= Truename;
	Data.ShortName		= ShortName;
	Data.AltName		= FileInfo.cAlternateFileName;
	Data.Description	= FileDescription;
	Data.Size			= FileSizeStr;
	Data.Attributes		= Attributes;
	Data.CreatedStamp	= CreatedStamp;
	Data.ModifiedStamp	= ModifiedStamp;
	Data.AccessedStamp	= AccessedStamp;
	Data.Owner			= Owner;
	Data.AssocStr		= AssocStr;
	Data.OpensWith		= OpensWith;
	Data.AssocType		= AssocType;
	Data.AttribsNumeric	= FileInfo.dwFileAttributes;

	if ( QueryIsFile( Truename ) )
		if ( !_wcsicmp( Ext, L".exe" ) || !_wcsicmp( Ext, L".dll" ) )
			GetFileVersion( Truename, Data.Version );


	//		Show the dialog box :

	DialogBoxParamW( MyDLLHandle, MAKEINTRESOURCE( IDD_DIALOG1 ), GetConsoleWindow(), DialogProc$, (LPARAM) &Data );
	return 0;
}



int PopupInfoKey( LPTSTR CommandLine, LPTSTR Current )
{
	if ( !CommandLine || !Current )
		return -1;


	wchar_t TempBuffer[MAX_FILENAME_LEN]	= L"";

	size_t index = Current - CommandLine;
	size_t i = 0, j = 0;
	wchar_t ch = '\0';
	BOOL quote = FALSE, infunc = FALSE;
	int infunc2 = 0;


	//		Parse the command line to find a filename :

	while ( CommandLine[i] != '\0' ) {
		ch = CommandLine[i++];

		if ( ch == '"' ) {
			quote = !quote;
			infunc = FALSE;
			continue;
		}

		if ( ch == '\0' )
			break;

		if ( infunc2 && !quote && ( ch == ']' ) ) {
			ch = ' ';
			infunc2--;
		}

		if ( !quote && ( ( ch == ',' ) || ( ch == ';' )  || ( ch == '<' )  || ( ch == '>' )  || ( ch == '|' )  || ( ch == '&' ) || iswspace(ch) ) ) {
			if ( i > index )
				break;

			j = 0;
			TempBuffer[j] = '\0';
			infunc = FALSE;
			continue;
		}

		if ( ( ch == '@' ) && ( j > 0 ) ) {
			if ( TempBuffer[j - 1] == '%' )
				infunc = TRUE;
		}
		else if ( ( ch == '[' ) && infunc ) {
			j = 0;
			TempBuffer[j] = '\0';
			infunc = FALSE;
			infunc2++;
			continue;
		}
		else if ( !__iscsym( ch ) )
			infunc = FALSE;

		if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
			TempBuffer[j++] = ch;
			TempBuffer[j] = '\0';
		}

	}


	//		Show the popup info window :

	return ShowFileInfo( TempBuffer );
}



DLLExports INT WINAPI key( LPKEYINFO24 ki )
{
	if ( ki->nKey == 0 ) {
		if ( !NewKeyHook )
			return 0;
		else if ( ki->pszKey == NULL )
			return 0;
	}


	int inKey = ki->nKey;
	int shiftKeys = 0;


	if ( NewKeyHook ) {
		if ( !inKey && ki->pszKey ) {
			inKey = ParseKeyName24( ki->pszKey );

			if ( !inKey )
				return 0;
		}
	}


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		shiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		shiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		shiftKeys |= 4;

	if ( shiftKeys == 3 )
		inKey |= 0x20000;
	else if ( shiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( shiftKeys == 7 ) )
		inKey |= 6000;


	if ( inKey != TriggerKey )
		return 0;

	PopupInfoKey( ki->pszLine, ki->pszCurrent );

	ki->nKey = 0;
	if ( NewKeyHook )
		ki->pszKey[0] = '\0';

	return 0;
}



int ParseZeroOneOption( LPTSTR inString )
{
	if ( !inString )
		return -1;

	int i = 0;

	if ( inString[i] == '/' )
		i++;

	while ( iswalpha( inString[i] ) )
		i++;

	if ( IsColonOrEquals( inString[i] ) )
		i++;

	if ( !wcscmp( inString + i, L"0" ) )
		return 0;
	else if ( !wcscmp( inString + i, L"1" ) )
		return 1;
	else if ( !_wcsicmp( inString + i, L"off" ) )
		return 0;
	else if ( !_wcsicmp( inString + i, L"on" ) )
		return 1;

	return -1;
}



int ParseTimeOption( LPTSTR inString )
{
	if ( !inString )
		return -1;

	int i = 0;

	if ( inString[i] == '/' )
		i++;

	while ( iswalpha( inString[i] ) )
		i++;

	if ( IsColonOrEquals( inString[i] ) )
		i++;

	if ( !wcscmp( inString + i, L"0" ) )
		return 0;
	else if ( !wcscmp( inString + i, L"1" ) )
		return 1;
	else if ( !wcscmp( inString + i, L"2" ) )
		return 2;
	else if ( !_wcsicmp( inString + i, L"off" ) )
		return 0;
	else if ( !_wcsicmp( inString + i, L"on" ) )
		return 1;
	else if ( !_wcsicmp( inString + i, L"utc" ) )
		return 2;

	return -1;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports INT WINAPI popinfo( LPTSTR String )
{
	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";

	int keyval					= 0;
	int rv						= 0;
	BOOL Save					= FALSE;
	BOOL Restore				= FALSE;
	BOOL List					= FALSE;
	BOOL Changes				= FALSE;
	unsigned int NumFilenames	= 0;
	int j						= 0;


	if ( !String )
		return -1;


	DivideCommandArgs( String );


	for ( int i = 0; i < MaxArgs; i++ ) {
		if ( arglen[i] ) {

			if ( arg[i][0] == '/' ) {

				switch ( arg[i][1] ) {

					case '?':

						return ShowCmdHelp( PopInfoHelpText, arg[i][2] == '?' );

					case 'S':

						Save = TRUE;
						break;

					case 'R':

						Restore = TRUE;
						break;

					case 'L':

						List = TRUE;
						break;

					case 'T':

						switch ( ParseTimeOption( arg[i] ) ) {

							case 0:
								LocalTimes = 0;
								break;

							case 1:
								LocalTimes = 1;
								break;

							case 2:
								LocalTimes = 2;
								break;

							default:
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"Unknown suboption \"%s\"\r\n", arg[i] );
								return 1;
						}

						Changes = TRUE;
						break;

					case 'C':

						switch ( ParseZeroOneOption( arg[i] ) ) {

							case 0:
								SubdirCounts = FALSE;
								break;

							case 1:
								SubdirCounts = TRUE;
								break;

							default:
								DisplayErrorHeader();
								Qprintf( ERRHANDLE, L"Unknown suboption \"%s\"\r\n", arg[i] );
								return 1;
						}

						Changes = TRUE;
						break;

					case 'F' :

						j = 2;

						if ( IsColonOrEquals( arg[i][j] ) )
							j++;

						if ( arg[i][j] == '\0' ) {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Missing filename:  \"%s\"\r\n", arg[i] );
							return 1;
						}

						NumFilenames++;
						break;

					default:
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
						return 1;
				}
			}
			else if ( keyval == 0 ) {
				keyval = ParseKeyName( arg[i] );

				if ( keyval == 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown key name \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( keyval == -1 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid hotkey \"%s\"\r\n", arg[i] );
					return 1;
				}
				else if ( keyval == -2 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Key not allowed \"%s\"\r\n", arg[i] );
					return 1;
				}

				Changes = TRUE;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( NumFilenames ) {
		if ( Save || Restore || List || Changes ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t combine /F with other options\r\n" );
			return 1;
		}

		rv = 0;

		for ( int i = 0; i < MaxArgs; i++ ) {
			if ( arglen[i] ) {
				if ( !wcsncmp( arg[i], L"/F", 2 ) ) {
					j = 2;
					if ( IsColonOrEquals( arg[i][j] ) )
						j++;

					if ( ShowFileInfo( arg[i] + j ) )
						rv = 2;
				}
			}
		}

		return rv;
	}


	if ( Save && Restore ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can't combine /R and /S.\r\n" );
		return 1;
	}

	if ( Restore && ( keyval != 0 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can't combine /R with a keyname.\r\n" );
		return 1;
	}

	if ( List ) {
		if ( Save || Restore || ( keyval != 0 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can't combine /L with other options.\r\n" );
			return 1;
		}

		ListLegalKeys();
		return 0;
	}


	if ( keyval ) {
		TriggerKey = keyval;
		GetKeyName( TriggerKey, shortbuf );
		Printf( L"PopInfo hotkey set to %s.\n", shortbuf );
	}

	if ( Save )
		if ( SaveOptions( TRUE ) )
			return 2;

	if ( Restore )
		if ( RestoreOptions( TRUE ) )
			return 2;

	if ( !keyval ) {
		GetKeyName( TriggerKey, shortbuf );
		Printf( L"PopInfo hotkey is %s.\n", shortbuf );
	}

	Printf( L"Smart timestamps %s, subdir counts %s\n", TimeOpts[LocalTimes], OnOff[SubdirCounts] );

	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR pszName, LPTSTR pszUsage )
{
	if ( !pszName || !pszUsage )
		return 0;


	if ( !_wcsicmp( pszName, L"popinfo" ) ) {
		wcscpy_s( pszUsage, SHORT_BUF_LEN, UsageText( PopInfoHelpText ) );
		return 1;
	}

	return 0;
}



DLLExports int WINAPI Help( LPTSTR pszName )
{
	if ( !pszName )
		return 0;


	if ( !_wcsicmp( pszName, L"popinfo" ) ) {
		MessageBox( GetForegroundWindow(), PopInfoHelpText, L"PopInfo", MB_OK | MB_ICONINFORMATION | MB_SETFOREGROUND );
		return -1;
	}

	return 0;
}
